==> couchbasedrupal-7.x-1.0.tar.gz: couchbasedrupal/src/Cache/CouchbaseBackend.php <==
<?php

namespace Drupal\couchbasedrupal\Cache;

/**
 * Couchbase implementation.
 */
class CouchbaseBackend extends CouchbaseBackendGeneric implements \DrupalCacheInterface {

  /**
   * Test info is loaded at database bootstrap phase
   * but this cache can be used earlier. Make sure
   * we have a valid test prefix before any operation on
   * the cache is performed.
   *
   * @var string
   */
  private static $test_run_id;

  /**
   * Couchbase bucket.
   *
   * @var \Drupal\couchbasedrupal\CouchbaseBucket
   */
  protected $bucket;

  /**
   * Name of current binary
   *
   * @var mixed
   */
  protected $bin;

  /**
   * Last time this cache was cleared.
   *
   * @var float
   */
  protected $lastClearTime = 0;

  /**
   * Add test prefix to current binary key.
   *
   * @param string $prefix
   */
  private static function AddTestPrefix(&$prefix) {
    // Try with the "official" test_run_id first.
    if (isset($GLOBALS['drupal_test_info']['test_run_id'])) {
      $prefix .= $GLOBALS['drupal_test_info']['test_run_id'];
      return;
    }
    // Try with the locally detected version.
    if (isset(static::$test_run_id)) {
      $prefix .= static::$test_run_id;
      return;
    }
  }

  /**
   * Get or Set the last clear time.
   *
   * @param float $time
   */
  protected function LastClear($time = NULL) {
    $key = 'BinaryLastCleared' . $this->binPrefix;
    if (empty($time)) {
      if ($cache = $this->bucket->get($key)) {
        $this->lastClearTime = $cache->value;
      }
      else {
        $this->bucket->insert($key, 0);
      }
      $this->lastClearTime = 0;
    }
    else {
      $this->bucket->upsert($key, $time);
      $this->lastClearTime = $time;
    }
  }

  /**
   * The constructor.
   *
   * @param string $bin
   *   The bin parameter.
   */
  public function __construct($bin) {
    $this->bucket = \Drupal\couchbasedrupal\CouchbaseManager::instance()->getBucketFromConfig('default', CouchbaseCacheTranscoder::class);
    $this->bin = $bin;
    $this->refreshRequestTime();
    $this->generateBinPrefix();
    $this->LastClear();
    $this->ensureIndexes();
  }

  /**
   * Generates the bin prefix.
   */
  protected function generateBinPrefix() {
    global $conf;
    // This method used to generate a key depending on $_SERVER['HOST']
    // but was refactored to align it to Memcache behaviour.
    // @see https://www.drupal.org/node/2426825
    $prefix = '';
    if ($prefix =(isset($conf['couchbase_key_prefix']) ? $conf['couchbase_key_prefix'] : FALSE)) {
      $prefix .= '|';
    }
    else {
      $prefix = (isset($_SERVER['HTTP_HOST']) ? $_SERVER['HTTP_HOST'] : '') . '|';
    }
    // When simpletest is running, emulate the simpletest database prefix here
    // to avoid the child site setting cache entries in the parent site (or the other way round).
    static::AddTestPrefix($prefix);
    $this->binPrefix = $prefix . $this->bin . '|';
  }

  /**
   * Function which retrieves a cached item.
   *
   * @param string $cid
   *   The cache id.
   *
   * @return string
   *   The cached item.
   */
  public function get($cid) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids);
    return reset($cache);
  }

  /**
   * Prepare a cached item.
   *
   * Checks that items are either permanent or did not expire, and unserializes
   * data as appropriate.
   *
   * @param mixed $cache
   *   An item loaded from cache_get() or cache_get_multiple().
   *
   * @return string
   *   The item with data unserialized as appropriate or FALSE if there is no
   *   valid item to load.
   */
  protected function prepareItem($cache) {

    $cache = $cache->value;
    if (!isset($cache->data)) {
      return FALSE;
    }

    if ($cache->created <= $this->lastClearTime) {
      return FALSE;
    }

    $cache->valid = $cache->expire == \CACHE_PERMANENT || $cache->expire >= $this->requestTime;

    if ($cache->valid === FALSE) {
      return FALSE;
    }

    return $cache;
  }

  /**
   * Get multiple cached items.
   *
   * @param array &$cids
   *   The array of keys.
   *
   * @return array
   *   The array of cached items.
   */
  public function getMultiple(&$cids) {
    // Translate the requested cache item IDs to Wincache keys.
    $map = array();
    foreach ($cids as $cid) {
      $map[$this->getBinKey($cid)] = $cid;
    }

    $result = $this->bucket->getMultiple(array_keys($map));
    $cache = array();
    foreach ($result as $key => $item) {
      $item = $this->prepareItem($item);
      if ($item) {
        $cache[$map[$key]] = $item;
      }
    }
    unset($result);

    $cids = array_diff($cids, array_keys($cache));
    return $cache;
  }

  /**
   * Garbage collection for get and getMultiple methods.
   */
  protected function garbageCollection() {
    // there is no such thing as garbageCollection for this implementation.
  }

  /**
   * Prepare a cache item.
   * */
  protected function prepareCacheItem($cid, $data, $expire, $headers) {
    $cache = new \stdClass();
    $cache->cid = $cid;
    $cache->created = REQUEST_TIME;
    $cache->created_microtime = round(microtime(TRUE), 3);
    $cache->expire = $expire;
    $cache->headers = $headers;
    $cache->data = $data;
    return $cache;
  }

  /**
   * Asign expiration to temporary items.
   *
   * @param int $expire
   */
  protected function fixTemporaryExpiration($expire) {
    // TODO: This should be grabbed from settings....
    // but that means bootstrapping up to "variable"
    // and that SUCKS and makes the implementation of this
    // way too complicated (see Memcache, Wincache and others...)
    if ($expire == \CACHE_TEMPORARY) {
      global $conf;
      $cache_lifetime = isset($conf['cache_lifetime']) ? $conf['cache_lifetime'] : 3600;
      // Add +20% random margin to temporary expirations.
      return $cache_lifetime + rand(0, $cache_lifetime * 0.20);
    }
    return $expire;
  }

  /**
   * Set an item in the cache.
   *
   * @param string $cid
   *   The key of the data to be cached.
   *
   * @param string $data
   *   The data to be cached.
   *
   * @param int $expire
   *   The expire mode.
   *
   * @param array $headers
   *   The headers.
   */
  public function set($cid, $data, $expire = CACHE_PERMANENT, array $headers = NULL) {

    $cache = $this->prepareCacheItem($cid, $data, $expire, $headers);

    if ($expire == \CACHE_PERMANENT) {
      return $this->bucket->upsert($this->getBinKey($cid), $cache);
    }

    $expire = $this->fixTemporaryExpiration($expire);
    $expire = $this->bucket->FixExpiration($expire);
    return $this->bucket->upsert($this->getBinKey($cid), $cache, array('expiry' => $expire));
  }

  /**
   * Adds a variable in user cache only if variable does not already exist in the cache.
   *
   * @param string $cid
   *   The key of the data to be cached.
   *
   * @param string $data
   *   The data to be cached.
   *
   * @param int $expire
   *   The expire mode.
   *
   * @param array $headers
   *   The headers.
   */
  public function add($cid, $data, $expire = \CACHE_PERMANENT, array $headers = NULL) {

    $cache = $this->prepareCacheItem($cid, $data, $expire, $headers);

    if ($expire == \CACHE_PERMANENT) {
      return $this->bucket->insert($this->getBinKey($cid), $cache);
    }

    $expire = $this->fixTemporaryExpiration($expire);
    $expire = $this->bucket->FixExpiration($expire);
    return $this->bucket->insert($this->getBinKey($cid), $cache, array('expiry' => $expire));
  }

  /**
   * Clear all cache items that were stored
   * using the CACHE_TEMPORARY mode.
   */
  public function clearTemporary() {
    $this->LastClear(microtime(TRUE));
  }

  /**
   * Remove an item from the cache.
   *
   * @param string $cid
   *   The key of the data to be removed.
   *
   * @param string $wildcard
   *   The wildcard.
   */
  public function clear($cid = NULL, $wildcard = FALSE) {
    if (empty($cid)) {
      // Get rid of all temporary items....
      $this->clearTemporary();
    }
    else {
      $keys = array();
      if ($wildcard) {
        $this->bucket->deleteAllByPrefix($this->getBinKey(''));
      }
      else {
        $keys = is_array($cid) ? $cid : [$cid];
        $keys = array_map(array($this, 'getBinKey'), $keys);
        $this->bucket->removeMultiple($keys);
      }
    }
  }

  /**
   * Checks cache is empty.
   *
   * @return bool
   *   TRUE if empty FALSE otherwise.
   */
  public function isEmpty() {
    return TRUE;
  }

  /**
   * Ensure that we have the proper indexes in place.
   */
  public function ensureIndexes() {

    static $verified;

    if (!isset($verified)) {
      if ($cache =  $this->bucket->get('couchbase_index_verified')) {
        $verified = $cache->value;
      }
    }

    if ($verified == TRUE) {
      return;
    }

    $bucket = $this->bucket;
    $bucket_name = $this->bucket->getName();

    try {
      $bucket->queryN1QL(\CouchbaseN1qlQuery::fromString("CREATE PRIMARY INDEX ON {$bucket_name} USING GSI", NULL, TRUE));
    }
    catch (\Exception $e) {
    }

    try {
      $bucket->queryN1QL(\CouchbaseN1qlQuery::fromString("CREATE INDEX id_ix on {$bucket_name}(meta().id);"), NULL, TRUE);
    }
    catch (\Exception $e) {
    }

    $this->bucket->upsert('couchbase_index_verified', TRUE, ['expiry' => 3600]);
    $verified = TRUE;

  }
}

==> couchbasedrupal-7.x-1.1.tar.gz: couchbasedrupal/src/Cache/CouchbaseBackend.php <==
<?php

namespace Drupal\couchbasedrupal\Cache;

/**
 * Couchbase implementation.
 */
class CouchbaseBackend extends CouchbaseBackendGeneric implements \DrupalCacheInterface {

  /**
   * Test info is loaded at database bootstrap phase
   * but this cache can be used earlier. Make sure
   * we have a valid test prefix before any operation on
   * the cache is performed.
   *
   * @var string
   */
  private static $test_run_id;

  /**
   * Couchbase bucket.
   *
   * @var \Drupal\couchbasedrupal\CouchbaseBucket
   */
  protected $bucket;

  /**
   * Name of current binary
   *
   * @var mixed
   */
  protected $bin;

  /**
   * Last time this cache was cleared.
   *
   * @var float
   */
  protected $lastClearTime = 0;

  /**
   * Cache lifetime
   *
   * @var int
   */
  protected $cache_lifetime;

  /**
   * Site prefix for couchbase keys
   *
   * @var string
   */
  protected $couchbase_key_prefix;

  /**
   * Add test prefix to current binary key.
   *
   * @param string $prefix
   */
  private static function AddTestPrefix(&$prefix) {
    // Try with the "official" test_run_id first.
    if (isset($GLOBALS['drupal_test_info']['test_run_id'])) {
      $prefix .= $GLOBALS['drupal_test_info']['test_run_id'];
      return;
    }
    // Try with the locally detected version.
    if (isset(static::$test_run_id)) {
      $prefix .= static::$test_run_id;
      return;
    }
  }

  /**
   * Get or Set the last clear time.
   *
   * @param float $time
   */
  protected function LastClear($time = NULL) {
    $key = 'BinaryLastCleared|' . $this->binPrefix;
    if ($time === NULL) {
      if ($cache = $this->bucket->get($key)) {
        $this->lastClearTime = $cache->value;
      }
      else {
        $this->bucket->insert($key, 0);
        $this->lastClearTime = 0;
      }
    }
    else {
      $this->bucket->upsert($key, $time);
      $this->lastClearTime = $time;
    }
  }

  /**
   * The constructor.
   *
   * @param string $bin
   *   The bin parameter.
   */
  public function __construct($bin) {
    $this->bucket = \Drupal\couchbasedrupal\CouchbaseManager::instance()->getBucketFromConfig('default', CouchbaseCacheTranscoder::class);
    $this->bin = $bin;
    $this->refreshRequestTime();
    $this->generateBinPrefix();
    $this->LastClear();
    $this->ensureIndexes();
    $this->loadCacheLifetime();
  }

  /**
   * Ensure availability of configuration options.
   */
  protected function loadCacheLifetime() {
    // If page_cache_without_database is enabled, we have to manually load the
    // $conf array out of cache_bootstrap.
    static $variables_loaded = FALSE;
    // NOTE: We don't call drupal_get_bootstrap_phase() because this would
    // break all 7.x Drupal installations prior to 7.33. For more information
    // see https://www.drupal.org/node/667098.
    if (!$variables_loaded && drupal_bootstrap(NULL, FALSE) < DRUPAL_BOOTSTRAP_VARIABLES) {
      // This cache can be used so early in the bootstrap that even testing environment has
      // not been initialized yet.... do so now otherwise caches between test environment
      // and regular environment can get mixed up.
      if ($test_prefix = drupal_valid_test_ua()) {
        // Keep a local copy of the current test_prefix.
        static::$test_run_id = $test_prefix;
      }
      // Now time to load the variables.
      global $conf;
      $variables_loaded = TRUE;
      // Try loading variables from cache. If that fails, we have to bootstrap
      // further in order to fetch them.
      if ($cached = cache_get('variables', 'cache_bootstrap')) {
        $variables = $cached->data;
        // Make sure variable overrides are applied, see variable_initialize().
        foreach ($conf as $name => $value) {
          $variables[$name] = $value;
        }
        $conf = $variables;
      }
      else {
        drupal_bootstrap(DRUPAL_BOOTSTRAP_VARIABLES, FALSE);
      }
    }
    $this->cache_lifetime = variable_get('cache_lifetime', 0);
    $this->couchbase_key_prefix = variable_get('couchbase_key_prefix', '');
  }

  /**
   * Generates the bin prefix.
   */
  protected function generateBinPrefix() {
    global $conf;
    // This method used to generate a key depending on $_SERVER['HOST']
    // but was refactored to align it to Memcache behaviour.
    // @see https://www.drupal.org/node/2426825
    $prefix = '';
    if (!empty($this->couchbase_key_prefix)) {
      $prefix .= '|';
    }
    else {
      $prefix = (isset($_SERVER['HTTP_HOST']) ? $_SERVER['HTTP_HOST'] : '') . '|';
    }
    // When simpletest is running, emulate the simpletest database prefix here
    // to avoid the child site setting cache entries in the parent site (or the other way round).
    static::AddTestPrefix($prefix);
    $this->binPrefix = $prefix . $this->bin . '|';
  }

  /**
   * Function which retrieves a cached item.
   *
   * @param string $cid
   *   The cache id.
   *
   * @return string
   *   The cached item.
   */
  public function get($cid) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids);
    return reset($cache);
  }

  /**
   * Prepare a cached item.
   *
   * Checks that items are either permanent or did not expire, and unserializes
   * data as appropriate.
   *
   * @param mixed $cache
   *   An item loaded from cache_get() or cache_get_multiple().
   *
   * @return string
   *   The item with data unserialized as appropriate or FALSE if there is no
   *   valid item to load.
   */
  protected function prepareItem($cache) {

    $cache = $cache->value;
    if (!isset($cache->data)) {
      return FALSE;
    }

    if ($cache->created <= $this->lastClearTime) {
      return FALSE;
    }

    $cache->valid = $cache->expire == \CACHE_PERMANENT || $cache->expire == \CACHE_TEMPORARY || $cache->expire >= $this->requestTime;

    if ($cache->valid === FALSE) {
      return FALSE;
    }

    return $cache;
  }

  /**
   * Get multiple cached items.
   *
   * @param array &$cids
   *   The array of keys.
   *
   * @return array
   *   The array of cached items.
   */
  public function getMultiple(&$cids) {
    // Translate the requested cache item IDs to Wincache keys.
    $map = array();
    foreach ($cids as $cid) {
      $map[$this->getBinKey($cid)] = $cid;
    }

    $result = $this->bucket->getMultiple(array_keys($map));
    $cache = array();
    foreach ($result as $key => $item) {
      $item = $this->prepareItem($item);
      if ($item) {
        $cache[$map[$key]] = $item;
      }
    }
    unset($result);

    $cids = array_diff($cids, array_keys($cache));
    return $cache;
  }

  /**
   * Garbage collection for get and getMultiple methods.
   */
  protected function garbageCollection() {
    // there is no such thing as garbageCollection for this implementation.
  }

  /**
   * Prepare a cache item.
   * */
  protected function prepareCacheItem($cid, $data, $expire, $headers) {
    $cache = new \stdClass();
    $cache->cid = $cid;
    $cache->created = REQUEST_TIME;
    $cache->created_microtime = round(microtime(TRUE), 3);
    $cache->expire = $expire;
    $cache->headers = $headers;
    $cache->data = $data;
    return $cache;
  }

  /**
   * Fix expirations for permanent and temporary items.
   *
   * @param int $expire
   */
  protected function fixExpirations($expire) {
    if ($expire == \CACHE_TEMPORARY) {
      // Add +20% random margin to temporary expirations.
      $expire = time() + $this->cache_lifetime + rand(0, $this->cache_lifetime * 0.2);
    }
    else if ($expire == \CACHE_PERMANENT) {
      // There is no such thing as a "permanent" item, or there is risk of stuff
      // staying endlessly in the cache... use 1 week +- 1 day
      $expire = time() + (3600 * 24 * 6) + rand(0, 3600 * 24 * 2);
    }
    // Now fix this for Couchbase.
    $expire = $this->bucket->FixExpiration($expire);
    return $expire;
  }

  /**
   * Set an item in the cache.
   *
   * @param string $cid
   *   The key of the data to be cached.
   *
   * @param string $data
   *   The data to be cached.
   *
   * @param int $expire
   *   The expire mode.
   *
   * @param array $headers
   *   The headers.
   */
  public function set($cid, $data, $expire = CACHE_PERMANENT, array $headers = NULL) {
    $cache = $this->prepareCacheItem($cid, $data, $expire, $headers);
    $expire = $this->fixExpirations($expire);
    return $this->bucket->upsert($this->getBinKey($cid), $cache, array('expiry' => $expire));
  }

  /**
   * Adds a variable in user cache only if variable does not already exist in the cache.
   *
   * @param string $cid
   *   The key of the data to be cached.
   *
   * @param string $data
   *   The data to be cached.
   *
   * @param int $expire
   *   The expire mode.
   *
   * @param array $headers
   *   The headers.
   */
  public function add($cid, $data, $expire = \CACHE_PERMANENT, array $headers = NULL) {
    $cache = $this->prepareCacheItem($cid, $data, $expire, $headers);
    $expire = $this->fixExpirations($expire);
    return $this->bucket->insert($this->getBinKey($cid), $cache, array('expiry' => $expire));
  }

  /**
   * Clear all cache items that were stored
   * using the CACHE_TEMPORARY mode.
   */
  public function clearTemporary() {
    $this->LastClear(microtime(TRUE));
  }

  /**
   * Remove an item from the cache.
   *
   * @param string $cid
   *   The key of the data to be removed.
   *
   * @param string $wildcard
   *   The wildcard.
   */
  public function clear($cid = NULL, $wildcard = FALSE) {
    if (empty($cid)) {
      // Get rid of all temporary items....
      $this->clearTemporary();
    }
    else {
      $keys = array();
      if ($wildcard) {
        $prefix = $cid == '*' ? '' : $cid;
        $this->bucket->deleteAllByPrefix($this->getBinKey($prefix));
      }
      else {
        $keys = is_array($cid) ? $cid : [$cid];
        $keys = array_map(array($this, 'getBinKey'), $keys);
        $this->bucket->removeMultiple($keys);
      }
    }
  }

  /**
   * Checks cache is empty.
   *
   * @return bool
   *   TRUE if empty FALSE otherwise.
   */
  public function isEmpty() {
    return TRUE;
  }

  /**
   * Ensure that we have the proper indexes in place.
   */
  public function ensureIndexes() {

    static $verified;

    if (!isset($verified)) {
      if ($cache =  $this->bucket->get('couchbase_index_verified')) {
        $verified = $cache->value;
      }
    }

    if ($verified == TRUE) {
      return;
    }

    $bucket = $this->bucket;
    $bucket_name = $this->bucket->getName();

    try {
      $bucket->queryN1QL(\CouchbaseN1qlQuery::fromString("CREATE PRIMARY INDEX ON {$bucket_name} USING GSI", NULL, TRUE));
    }
    catch (\Exception $e) {
    }

    try {
      $bucket->queryN1QL(\CouchbaseN1qlQuery::fromString("CREATE INDEX id_ix on {$bucket_name}(meta().id);"), NULL, TRUE);
    }
    catch (\Exception $e) {
    }

    $this->bucket->upsert('couchbase_index_verified', TRUE, ['expiry' => 3600]);
    $verified = TRUE;

  }
}
